#include "copter.h"
#include <SDL.h>
#include <sstream>

#define COPTER_KEY_YAW_LEFT 	SDLK_e
#define COPTER_KEY_YAW_RIGHT	SDLK_d
#define COPTER_KEY_LEFT 	SDLK_LEFT
#define COPTER_KEY_RIGHT 	SDLK_RIGHT
#define COPTER_KEY_FORWARD 	SDLK_UP
#define COPTER_KEY_BACK 	SDLK_DOWN
#define COPTER_KEY_UP 		SDLK_q
#define COPTER_KEY_DOWN 	SDLK_a

#define MainRotorAngleX 10.0
#define MainRotorAngleZ 10.0

#define CriticalHeight 100.0
#define TopHeight 5000.0
#define FallingSpeed 1.0

Copter::Copter(MetaMesh *meshes, Vector3D _position, Landscape *l):Mobile(meshes,_position,l)
{
        CollectiveSteer=new Steer(1.0);
        RearRotorSteer=new Steer(1.0);
        LeftRightSteer=new Steer(1.0);
        ForwardBackSteer=new Steer(1.0);

        MainRotorSpeed=1000.0;
        RearRotorSpeed=1000.0;

        RollSpeed=12.0;
        RollMax=50.0;
        PitchSpeed=20.0;
        PitchMax=80.0;

        ForwardBackSpeed=40.0;
        LeftRightSpeed=-5.0;
        LeftRightYawSpeed=0.5;
        YawSpeed=45.0;

        CollectiveSpeed=400.0;

        Friction=Vector3D(0.80,0.50,0.10);

        CentrifugalRoll=-0.3;
        CentrifugalPitch=-0.3;

        Height=0.0;
        MainRotorPhase=0.0;
        RearRotorPhase=0.0;
}


void Copter::SteerStep(Uint8 *keys, int k1, int k2, Steer *s, double t)
{
        double d;
        if (keys[k1]==SDL_PRESSED)
                d=-1.0;
        else
                if (keys[k2]==SDL_PRESSED)
                        d=1.0;
                else
                        d=0.0;
        s->Step(d,t);
}

void Copter::CalcAcceleration()
{
        Vector3D a1,a2,a3,a4;
        Vector3D f;
        a1=Vector3D(0,CollectiveSteer->Get()*CollectiveSpeed,0);
        a2=Vector3D(Roll*LeftRightSpeed,0,Pitch*ForwardBackSpeed);
        a2=Orientation*a2;
        f=Orientation*Friction;
        f.x=fabs(f.x);
        f.y=fabs(f.y);
        f.z=fabs(f.z);
        a3.x=-Velocity.x*f.x;
        a3.y=-Velocity.y*f.y;
        a3.z=-Velocity.z*f.z;
        if (Position.y>TopHeight)
                a4=Vector3D(0,-(Position.y-TopHeight)*FallingSpeed,0);
        Acceleration=a1+a2+a3+a4;
}

void Copter::RotorsStep(double t)
{
        Matrix4 mx,my,mz;

        MainRotorPhase+=MainRotorSpeed*t;
        mx.CreateRotateXMatrix(MainRotorAngleZ*ForwardBackSteer->Get());
        my.CreateRotateYMatrix(MainRotorPhase);
        mz.CreateRotateZMatrix(MainRotorAngleX*LeftRightSteer->Get());
        SetObjectMatrix("MainRotor",mx*mz*my);

        RearRotorPhase+=RearRotorSpeed*t;
        mx.CreateRotateXMatrix(MainRotorPhase);
        SetObjectMatrix("RearRotor",mx);
}

void Copter::StepYaw(double t)
{
        Matrix4 r2,r3,m;
        Vector3D v1,v2;
        r2.CreateRotateZMatrix(Roll);
        r3.CreateRotateXMatrix(Pitch);
        m=r3*r2;
        v1=Vector3D(RearRotorSteer->Get()*YawSpeed*t+Roll*LeftRightYawSpeed*t,0,0);
        v2=m*v1;
        Yaw+=v2.x;
        Roll+=v2.z;
        Pitch-=v2.y;
        if (Yaw>360)
                Yaw-=360;
        if (Yaw<0)
                Yaw+=360;
}

void Copter::DrawInfoColumn(double x, double y, double w, double h, int steps, double value)
{
        double delta_y,current_y,delta_x;
        int i;
        delta_y=h/(steps+1);
        delta_x=w/10;
        current_y=y;
        glBegin(GL_LINES);
        glColor3f(0.7,0.7,0.7);
        glVertex2f(x,y);
        glVertex2f(x+w,y);
        glVertex2f(x,y+h);
        glVertex2f(x+w,y+h);
        for (i=0;i<steps;i++)
        {
                current_y+=delta_y;
                glVertex2f(x+delta_x,current_y);
                glVertex2f(x+w-delta_x,current_y);
        }
        glColor3f(1,1,1);
        glVertex2f(x,y+h/2+value*h/2);
        glVertex2f(x+w,y+h/2+value*h/2);
        glEnd();
}

void Copter::DrawInfo(Text2D *text2d)
{
        ostringstream oss;
        DrawInfoColumn(0.3,0.05,0.015,0.15,6,CollectiveSteer->Get());
        text2d->Print(0.25,0.00,"collective");
        DrawInfoColumn(0.4,0.05,0.015,0.15,6,RearRotorSteer->Get());
        text2d->Print(0.35,0.00,"rear rotor");
        DrawInfoColumn(0.5,0.05,0.015,0.15,6,LeftRightSteer->Get());
        text2d->Print(0.45,0.00,"left/right");
        DrawInfoColumn(0.6,0.05,0.015,0.15,6,ForwardBackSteer->Get());
        text2d->Print(0.55,0.00,"forward/back");

        oss.setf(ios::fixed);
        oss.precision(0);

        oss.str("");
        oss<<"absolute height: "<<Position.y;
        glColor3f(1,1,1);
        text2d->Print(0.01,0.975,oss.str());

        oss.str("");
        if (Height>CriticalHeight)
                glColor3f(1,1,1);
        else
                glColor3f(1,0,0);
        oss<<"relative height: "<<Height;
        text2d->Print(0.01,0.950,oss.str());

        oss.str("");
        oss<<"velocity: "<<Velocity.Length();
        glColor3f(1,1,1);
        text2d->Print(0.01,0.925,oss.str());
}

void Copter::Step(double t, Uint8 *keys)
{
        SteerStep(keys,COPTER_KEY_YAW_RIGHT,COPTER_KEY_YAW_LEFT,RearRotorSteer,t);
        SteerStep(keys,COPTER_KEY_RIGHT,COPTER_KEY_LEFT,LeftRightSteer,t);
        SteerStep(keys,COPTER_KEY_FORWARD,COPTER_KEY_BACK,ForwardBackSteer,t);
        SteerStep(keys,COPTER_KEY_DOWN,COPTER_KEY_UP,CollectiveSteer,t);

        StepYaw(t);

        Roll+=LeftRightSteer->Get()*RollSpeed*t+CentrifugalRoll*Roll*t;
        if (Roll>RollMax)
                Roll=RollMax;
        if (Roll<-RollMax)
                Roll=-RollMax;

        Pitch+=ForwardBackSteer->Get()*PitchSpeed*t+CentrifugalPitch*Pitch*t;
        if (Pitch>PitchMax)
                Pitch=PitchMax;
        if (Pitch<-PitchMax)
                Pitch=-PitchMax;

        CalcAcceleration();

        Position+=Velocity*t+Acceleration*t*t/2.0;
        Position.WrapXZ(landscape->GetSizeX(),landscape->GetSizeZ());
        Velocity+=Acceleration*t;

        RotorsStep(t);
        SetMatrices();

        CheckCrash();
}
