#include "landscape.h"
#include "landscapetexture.h"
#include <fstream>
#include <iostream>
#include <cmath>
#include <string>
#include <sstream>
#include "image.h"
#include "common.h"

#define MaxMerges 500
#define MaxSplits 500

Landscape::Landscape(string filename, int _PatchLevel, double _ViewDistance, int _TextureSize, int _MaxTriangles)
{
        cout <<"---\nnew Landscape\n";
        LoadHeightMap(filename);

        SizeX=hm->GetSizeX();
        SizeZ=hm->GetSizeZ();
        TextureSize=_TextureSize;

        MaxTriangles=_MaxTriangles;
        MaxQsSize=MaxTriangles+100;
        MaxQmSize=MaxTriangles/4;
        BTTPoolSize=MaxTriangles*2;
        ViewDistance=_ViewDistance;

        indices=new GLuint[MaxIndices];

        bp=new BTTPool(BTTPoolSize);
        Qs=new SplitQueue(MaxQsSize, SizeX, SizeZ, ViewDistance);
        Qm=new MergeQueue(MaxQmSize, SizeX, SizeZ, ViewDistance);

        PatchLevel=_PatchLevel;
        PatchSize=hm->GetScaleXZ()*pow(2.0,PatchLevel/2.0);
        PatchesX=(int)(SizeX/PatchSize);
        PatchesZ=(int)(SizeZ/PatchSize);

        patches=NULL;

        LightDirection=-LightOrigin;
        LightDistance=LightDirection.Length();
        LightDirection.Normalize();

        cout <<"new Landscape - done\n---\n";
}

Landscape::~Landscape()
{
        int i,j;
        Vector3D v;
        v=Vector3D(0,0,0);
        cout << "\n---\n [ delete Landscape ]\n";
        while (!Qm->Empty())
        {
                MergePhase(&v);
        }
        if (patches!=NULL)
        {
                for (i=0;i<PatchesX;i++)
                {
                        for (j=0;j<PatchesZ;j++)
                        {
                                delete patches[i][j];
                        }
                        delete [] patches[i];
                }
                delete [] patches;
        }
        delete Qs;
        delete Qm;
        delete bp;
        cout << " [ delete Landscape - done ]\n---\n";
}

void Landscape::CreatePatches(string LocalDataDirectory)
{
        int i,j;
        Patch *ln, *rn, *tn, *bn;
        GLuint texture;
        int PatchSizeInPoints,NumberOfPoints,InfoTreeSize;
        Image *image;

        PatchSizeInPoints=(int)pow(2.0,PatchLevel/2.0);
        NumberOfPoints=PatchSizeInPoints+1;
        InfoTreeSize=(int)pow(2.0,(double)PatchLevel-1)-1;

        LeftIndicesTree=new IndicesTree(PatchLevel, NumberOfPoints, PatchSizeInPoints,PatchSizeInPoints,0,0,0,PatchSizeInPoints);
        RightIndicesTree=new IndicesTree(PatchLevel, NumberOfPoints, 0,0,PatchSizeInPoints,PatchSizeInPoints,PatchSizeInPoints,0);

        cout <<"Creating "<<PatchesX<<"x"<<PatchesZ<<" patches...\n";

        patches=new Patch ** [PatchesX];
        for (i=0;i<PatchesX;i++)
        {
                patches[i]=new Patch * [PatchesZ];
                for (j=0;j<PatchesZ;j++)
                {
                        image=new Image(CreateTextureFileName(LocalDataDirectory,i,j));
                        texture=image->GenerateTexture();
                        delete image;
                        patches[i][j]=new Patch(i*PatchSize,j*PatchSize,(i+1)*PatchSize,(j+1)*PatchSize,PatchLevel,hm,bp,TextureSize, LeftIndicesTree, RightIndicesTree,indices);
                        patches[i][j]->SetTexture(texture);
                }
        }

        cout <<(sizeof(VertexStruct)+sizeof(TextureStruct))*NumberOfPoints*NumberOfPoints*PatchesX*PatchesZ<<" bytes allocated for Patch struct arrays\n";
        cout <<InfoTreeSize*2*PatchesX*PatchesZ<<" bytes allocated for TriangleInfoTree error arrays\n";

        for (i=0;i<PatchesX;i++)
        {
                for (j=0;j<PatchesZ;j++)
                {
                        if (i>0)
                                ln=patches[i-1][j];
                        else
                                ln=patches[PatchesX-1][j];
                        if (i<PatchesX-1)
                                rn=patches[i+1][j];
                        else
                                rn=patches[0][j];
                        if (j>0)
                                tn=patches[i][j-1];
                        else
                                tn=patches[i][PatchesZ-1];
                        if (j<PatchesZ-1)
                                bn=patches[i][j+1];
                        else
                                bn=patches[i][0];
                        patches[i][j]->SetNeighbours(ln,rn,tn,bn);
                }
        }
        for (i=0;i<PatchesX;i++)
        {
                for (j=0;j<PatchesZ;j++)
                {
                        patches[i][j]->FillQs(Qs);
                }
        }
}

void Landscape::LoadHeightMap(string filename)
{
        ifstream *f;
        double ScaleXZ,ScaleY;
        string mapname;
        double x,y,z;

        f=new ifstream(filename.c_str());
        Expect(f,"map");
        (*f)>>mapname;
        Expect(f,"LandscapeScaleXZ");
        (*f)>>ScaleXZ;
        Expect(f,"LandscapeScaleY");
        (*f)>>ScaleY;
        Expect(f,"Light");
        (*f)>>x;
        (*f)>>y;
        (*f)>>z;
        LightOrigin=Vector3D(x,y,z);
        f->close();
        delete f;

        hm=new HeightMap(mapname,ScaleXZ,ScaleY);
}

int Landscape::CalcXInPatches(int _x)
{
        int x;
        x=_x;
        if (x<0)
                x=x-PatchesX*(x/PatchesX-1);
        x=x%PatchesX;
        return x;
}

int Landscape::CalcZInPatches(int _z)
{
        int z;
        z=_z;
        if (z<0)
                z=z-PatchesZ*(z/PatchesZ-1);
        z=z%PatchesZ;
        return z;
}

void Landscape::DrawPatches(int x1, int z1, int x2, int z2, GLuint primitive)
{
        int nindices,n;
        int enabled;
        glEnableClientState(GL_VERTEX_ARRAY);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        glEnable(GL_TEXTURE_2D);
        glEnable (GL_DEPTH_TEST);
        glEnable (GL_CULL_FACE);

        int i,j;
        int x,z;
        frustum.Calc();
        enabled=0;
        nindices=0;
        for (i=0;i<PatchesX;i++)
                for (j=0;j<PatchesZ;j++)
                        patches[i][j]->Disable();
        for (i=x1;i<x2;i++)
                for (j=z1;j<z2;j++)
                {
                        x=CalcXInPatches(i);
                        z=CalcZInPatches(j);
                        if (patches[x][z]->CheckIfEnabled(i*PatchSize,j*PatchSize,&frustum))
                        {
                                enabled++;
                                n=patches[x][z]->GenerateIndices();
                                nindices+=n;
                                glMatrixMode(GL_MODELVIEW);
                                glPushMatrix();
                                glTranslatef(PatchSize*i,0,PatchSize*j);
                                patches[x][z]->Draw(primitive);
                                glMatrixMode(GL_MODELVIEW);
                                glPopMatrix();
                        }
                }
        glDisableClientState(GL_VERTEX_ARRAY);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        glDisable(GL_TEXTURE_2D);
        glDisable (GL_DEPTH_TEST);
        glDisable (GL_CULL_FACE);
}

void Landscape::Draw(Vector3D position,GLuint primitive)
{
        int x1, z1, x2, z2;
        int x,z,r;

        x=(int)(position.x/(PatchSize));
        z=(int)(position.z/(PatchSize));
        r=2+(int)(ViewDistance/PatchSize);
        x1=x-r;
        x2=x+r;
        z1=z-r;
        z2=z+r;

        DrawPatches(x1,z1,x2,z2, primitive);

}

int Landscape::GetPatchesX()
{
        return PatchesX;
}

int Landscape::GetPatchesZ()
{
        return PatchesZ;
}

void Landscape::MergePhase(Vector3D *CameraEye)
{
        MergeableDiamond *d;
        BTTNode *t1, *t2;
        TriangleInfoTree *vt1, *vt2;
        int n1, n2;
        LandscapeTriangle *t;

        if (Qm->Empty())
        {
                cout <<"Landscape::MergePhase: nothing to merge as yet\n";
                exit(-1);
        }
        d=Qm->GetHighest();

        t1=d->FirstBTTNode;
        t2=d->SecondBTTNode;

        Qs->RemoveBTTNodeFromQs(t1->LeftChild);
        Qs->RemoveBTTNodeFromQs(t1->RightChild);
        Qs->RemoveBTTNodeFromQs(t2->LeftChild);
        Qs->RemoveBTTNodeFromQs(t2->RightChild);

        t1->Merge(bp);

        vt1=d->First_it;
        n1=d->First_it_node;

        Qm->RemoveBTTNodeFromQm(t1);

        t=Qs->AllocateTriangle(t1,vt1,n1);
        Qs->Add(t,CameraEye);

        t2->Merge(bp);

        vt2=d->Second_it;
        n2=d->Second_it_node;

        t=Qs->AllocateTriangle(t2,vt2,n2);
        Qs->Add(t,CameraEye);

        if (t1->Parent)
                AddNewMergeableDiamond(t1->Parent,CameraEye);
        if (t2->Parent)
                AddNewMergeableDiamond(t2->Parent,CameraEye);

}

void Landscape::AddNewMergeableDiamond(BTTNode *tree, Vector3D *CameraEye)
{
        BTTNode *tree_bn;
        MergeableDiamond *d;
        LandscapeTriangle *t1,*t2;

        tree_bn=tree->BaseNeighbour;
        /** check if base neighbour has children */
        if (!tree_bn->LeftChild)
                return;

        if ( (tree->level==tree_bn->level) &&
                        (!tree_bn->LeftChild->LeftChild) &&
                        (!tree_bn->RightChild->LeftChild) &&
                        (!tree->LeftChild->LeftChild) &&
                        (!tree->RightChild->LeftChild)  )
        {
                t1=Qs->GetTriangle(tree->LeftChild);
                t2=Qs->GetTriangle(tree_bn->LeftChild);
                d=Qm->AllocateDiamond(tree,tree_bn,t1->First_it,t1->First_it_node>>1, t2->First_it,
                                      t2->First_it_node>>1);
                Qm->Add(d,CameraEye);
        }
}

void Landscape::SplitPhase(Vector3D *CameraEye)
{
        ForceSplit(Qs->GetHighest()->FirstBTTNode,CameraEye);
}

void Landscape::ForceSplit(BTTNode *tree,Vector3D *CameraEye)
{
        if (tree->BaseNeighbour->level!=tree->level)
                ForceSplit(tree->BaseNeighbour, CameraEye);

        BTTNode *tree_bn;
        LandscapeTriangle *lt, *lt_bn;

        tree_bn=tree->BaseNeighbour;
        lt=Qs->GetTriangle(tree);
        lt_bn=Qs->GetTriangle(tree_bn);

        tree->SplitTwo(bp);

        AddSplitedTriangleChildren(lt,CameraEye);

        AddSplitedTriangleChildren(lt_bn,CameraEye);

        AddNewMergeableDiamond(tree,CameraEye);
}

void Landscape::AddSplitedTriangleChildren(LandscapeTriangle *lt, Vector3D *CameraEye)
{
        LandscapeTriangle *nt;
        TriangleInfoTree *vt;
        BTTNode *tree;
        BTTNode *parent;

        vt=lt->First_it;
        tree=lt->FirstBTTNode;
        parent=tree->Parent;

        nt=Qs->AllocateTriangle(tree->LeftChild,vt,lt->First_it_node<<1);
        Qs->Add(nt,CameraEye);
        nt=Qs->AllocateTriangle(tree->RightChild,vt,1+(lt->First_it_node<<1));
        Qs->Add(nt,CameraEye);
        Qs->RemoveBTTNodeFromQs(tree);
        if (parent)
                Qm->RemoveBTTNodeFromQm(parent);

}

void Landscape::Update(Vector3D *CameraEye, double TargetPriority)
{
        Qs->RecalcPriorities(CameraEye);
        Qm->RecalcPriorities(CameraEye);
        while ((TargetPriority>1-Qm->GetHighestPriority()))
        {
                MergePhase(CameraEye);
        }
        while (((Qs->GetHighestPriority()>TargetPriority) && (Qs->GetSize()<MaxTriangles)))
        {
                SplitPhase(CameraEye);
        }
}

string Landscape::GetQueuesSizeString()
{
        ostringstream oss;
        oss.str("");
        oss<<"Qs: "<<Qs->GetSize()<<" Qm: "<<Qm->GetSize();
        return oss.str();
}

double Landscape::GetHeight(double x, double z)
{
        return hm->GetHeight(x,z);
}

#define MaxLandscapeTextures 20

void Landscape::GenerateTextures(string filename, string LocalDataDirectory, string TexturesDirectory, double LightAmbient, double LightEnergy, int MaxShadowLength)
{
        int i,j,done;
        ifstream *f;
        string texturename;
        int ntextures;

        Vector3D ShadowStep;
        double LightHeight;

        LightHeight=LightOrigin.y;
        ShadowStep=-LightOrigin;
        ShadowStep.y=0.0;
        ShadowStep.Normalize();
        ShadowStep=ShadowStep*hm->GetScaleXZ();
        hm->SetLight(LightDirection,ShadowStep,LightHeight,LightDistance,MaxShadowLength);

        cout <<"reading textures from "<<TexturesDirectory<<"\n";
        LandscapeTexture *textures[MaxLandscapeTextures];
        f=new ifstream(filename.c_str());
        ntextures=0;
        done=0;
        while (!done)
        {
                (*f)>>texturename;
                if (!f->eof())
                {
                        textures[ntextures]=new LandscapeTexture(f,TexturesDirectory+texturename);

                        ntextures++;
                }
                else
                        done=1;
        }
        cout <<ntextures<<" textures\n";
        for (i=0;i<PatchesX;i++)
        {
                for (j=0;j<PatchesZ;j++)
                {
                        GenerateTexture(i*PatchSize,j*PatchSize,(i+1)*PatchSize,(j+1)*PatchSize,textures,ntextures,CreateTextureFileName(LocalDataDirectory,i,j), LightAmbient, LightEnergy);
                }
                cout <<(100*(i+1)/(PatchesX))<<"% done\n";
        }
        for (i=0;i<ntextures;i++)
        {
                delete textures[i];
        }
}

void Landscape::GenerateTexture(double x1, double z1, double x2, double z2, LandscapeTexture **textures, int ntextures, string filename, double LightAmbient, double LightEnergy)
{
        SDL_Surface *surface;
        int i,j,k;
        double sumr,sumg,sumb;
        Uint8 r,g,b;
        double _r,_g,_b;
        double sum,w;
        double ax,bx,az,bz;
        double x,z,h;
        double nx,nz,n;
        double light;
        Uint8 *ptr,*lineptr;
        double ScaleY;

        ScaleY=hm->GetScaleY();

        surface=SDL_CreateRGBSurface(SDL_SWSURFACE, TextureSize, TextureSize,24, 255<<16, 255<<8, 255, 0);
        if (surface==NULL)
        {
                cout <<"error SDL_CreateRGBSurface\n";
                exit(-1);
        }

        ax=(x2-x1)/(TextureSize-2);
        az=(z2-z1)/(TextureSize-2);
        bx=x1-ax;
        bz=z1-az;

        lineptr=(Uint8*)surface->pixels;
        for (i=0;i<TextureSize;i++)
        {
                ptr=lineptr;
                for (j=0;j<TextureSize;j++)
                {
                        x=ax*i+bx;
                        z=az*j+bz;
                        if (x<0.0)
                                x+=SizeX;
                        if (z<0.0)
                                z+=SizeZ;
                        if (x>=SizeX)
                                x-=SizeX;
                        if (z>=SizeZ)
                                z-=SizeZ;

                        hm->GetHeightAndGradient(x,z,&h,&nx,&nz);

                        if (fabs(nx)>fabs(nz))
                                n=fabs(nx);
                        else
                                n=fabs(nz);

                        light=LightAmbient+LightEnergy*hm->Intersect(Vector3D(x,h,z),nx,nz);

                        sumr=0;
                        sumg=0;
                        sumb=0;
                        sum=0.0;
                        for (k=0;k<ntextures;k++)
                        {
                                w=textures[k]->GetPixel(x/SizeX,z/SizeZ,h/ScaleY,atan(n)/(M_PI/2.0),&r,&g,&b);
                                sum+=w;
                                sumr+=r*w;
                                sumg+=g*w;
                                sumb+=b*w;
                        }
                        if (sum>0.0)
                        {
                                _r=light*sumr/sum;
                                _g=light*sumg/sum;
                                _b=light*sumb/sum;
                                if (_r>255)
                                        r=255;
                                else
                                        r=(Uint8)_r;
                                if (_g>255)
                                        g=255;
                                else
                                        g=(Uint8)_g;
                                if (_b>255)
                                        b=255;
                                else
                                        b=(Uint8)_b;

                        }
                        else
                        {	r=0;
                                g=0;
                                b=0;
                        }

                        *ptr=r;
                        ptr++;
                        *ptr=g;
                        ptr++;
                        *ptr=b;
                        ptr++;
                }
                lineptr+=surface->pitch;
        }
        SDL_SaveBMP(surface,filename.c_str());
        SDL_FreeSurface(surface);
}

string Landscape::CreateTextureFileName(string dir, int x, int z)
{
        ostringstream oss;
        oss.str("");
        oss<<dir<<"/"<<"patch_"<<x<<"_"<<z<<".bmp";
        return oss.str();
}

double Landscape::GetSizeX()
{
        return SizeX;
}

double Landscape::GetSizeZ()
{
        return SizeZ;
}

Vector3D Landscape::GetLightDirection()
{
        return LightDirection;
}
