#include <cmath>
#include "matrix4.h"
#include <GL/gl.h>
#include <assert.h>
#include <stdio.h>

Matrix4::Matrix4 ()
{
        Zero ();
}

Matrix4::Matrix4 (const Matrix4 &m)
{
        int i,j;
        for (i = 0; i < 4; i++)
                for (j = 0; j < 4; j++)
                        e[i][j]=m.e[i][j];
}

Matrix4::Matrix4 (Vector3D &v)
{
        Zero();
        e[0][0]=v.x;
        e[1][1]=v.y;
        e[2][2]=v.z;
}

Matrix4::~Matrix4 ()
{
}

void Matrix4::Zero ()
{
        int i, j;
        for (i = 0; i < 4; i++)
                for (j = 0; j < 4; j++)
                        e[i][j] = 0.0;
}

void Matrix4::LoadIdentity()
{
        Zero();
        e[0][0]=1;
        e[1][1]=1;
        e[2][2]=1;
        e[3][3]=1;
}

void Matrix4::CreateRotateXMatrix (double angle)
{
        Zero ();
        angle = M_PI * angle / 180;
        e[0][0] = 1.0;
        e[1][1] = cos (angle);
        e[1][2] = -sin (angle);
        e[2][1] = sin (angle);
        e[2][2] = cos (angle);
        e[3][3] = 1.0;
}

void Matrix4::CreateRotateYMatrix (double angle)
{
        Zero ();
        angle = M_PI * angle / 180;
        e[0][0] = cos (angle);
        e[0][2] = sin (angle);
        e[1][1] = 1.0;
        e[2][0] = -sin (angle);
        e[2][2] = cos (angle);
        e[3][3] = 1.0;
}

void Matrix4::CreateRotateZMatrix (double angle)
{
        Zero ();
        angle = M_PI * angle / 180;
        e[0][0] = cos (angle);
        e[0][1] = -sin (angle);
        e[1][0] = sin (angle);
        e[1][1] = cos (angle);
        e[2][2] = 1.0;
        e[3][3] = 1.0;
}

void Matrix4::CreateTranslateMatrix (Vector3D v)
{
        Zero ();
        e[0][0] = 1;
        e[1][1] = 1;
        e[2][2] = 1;
        e[0][3] = v.x;
        e[1][3] = v.y;
        e[2][3] = v.z;
        e[3][3] = 1;
}

void Matrix4::CreateScaleMatrix (Vector3D v)
{
        Zero();
        e[0][0]=v.x;
        e[1][1]=v.y;
        e[2][2]=v.z;
        e[3][3]=1;
}

Vector3D Matrix4::operator * (Vector3D v1)
{
        Vector3D v;
        double x, y, z;
        x = v1.x;
        y = v1.y;
        z = v1.z;
        v=Vector3D(e[0][0] * x + e[0][1] * y + e[0][2] * z +e[0][3],
                   e[1][0] * x + e[1][1] * y + e[1][2] * z +e[1][3],
                   e[2][0] * x + e[2][1] * y + e[2][2] * z +e[2][3]);
        return v;
}

Vector3D operator * (Vector3D v1,Matrix4 m)
{
        Vector3D v;
        double x, y, z;
        x = v1.x;
        y = v1.y;
        z = v1.z;
        v=Vector3D(m.e[0][0] * x + m.e[1][0] * y + m.e[2][0] * z+m.e[3][0],
                   m.e[0][1] * x + m.e[1][1] * y + m.e[2][1] * z+m.e[3][1],
                   m.e[0][2] * x + m.e[1][2] * y + m.e[2][2] * z+m.e[3][2]);
        return v;
}

Matrix4 Matrix4::operator *(Matrix4 m)
{
        int i, j, k;
        Matrix4 matrix;
        matrix.Zero ();
        for (i = 0; i < 4; i++)
                for (j = 0; j < 4; j++)
                        for (k = 0; k < 4; k++)
                                matrix.e[i][j] += e[i][k] * m.e[k][j];
        return matrix;
}

Matrix4 Matrix4::operator +(Matrix4 m)
{
        int i, j;
        Matrix4 matrix;
        matrix.Zero ();
        for (i = 0; i < 4; i++)
                for (j = 0; j < 4; j++)
                        matrix.e[i][j] = e[i][j] + m.e[i][j];
        return matrix;
}

void Matrix4::operator *= (Matrix4 m)
{
        int i, j, k;
        Matrix4 matrix;
        matrix.Zero ();
        for (i = 0; i < 4; i++)
                for (j = 0; j < 4; j++)
                        for (k = 0; k < 4; k++)
                                matrix.e[i][j] += e[i][k] * m.e[k][j];
        *this=matrix;
}

Matrix4 Matrix4::operator *(double a)
{
        Matrix4 m;
        int i,j;
        for (i = 0; i < 4; i++)
                for (j = 0; j < 4; j++)
                        m.e[i][j]=e[i][j]*a;
        return m;
}

Matrix4 Matrix4::operator /(double a)
{
        Matrix4 m;
        int i,j;
        assert(a!=0.0);
        for (i = 0; i < 4; i++)
                for (j = 0; j < 4; j++)
                        m.e[i][j]=e[i][j]/a;
        return m;
}


void Matrix4::CreateRotateMatrix (Vector3D v)
{
        Matrix4 mx,my,mz,m;
        mx.CreateRotateXMatrix (v.x);
        my.CreateRotateYMatrix (v.y);
        mz.CreateRotateZMatrix (v.z);
        m=mz * my * mx;
        *this=m;
}

void Matrix4::Print ()
{
        int i, j;
        for (i = 0; i < 4; i++)
        {
                for (j = 0; j < 4; j++)
                        printf ("%.3f ", e[i][j]);
                printf ("\n");
        }
}

Matrix4 Matrix4::Transpose()
{
        int i,j;
        Matrix4 m;
        for (i=0;i<4;i++)
                for (j=0;j<4;j++)
                        m.e[i][j]=e[j][i];
        return m;
}

double Matrix4::Cofactor(int x, int y)
{
        int xx[3];
        int yy[3];
        int i,j;
        double det;
        double plus, minus;
        for (i=0;i<3;i++)
        {
                if (i<x)
                        xx[i]=i;
                else
                        xx[i]=i+1;
                if (i<y)
                        yy[i]=i;
                else
                        yy[i]=i+1;
        }
        det=0;
        for (i=0;i<3;i++)
        {
                plus=1;
                minus=1;
                for (j=0;j<3;j++)
                {
                        plus=plus*e[xx[j]][yy[(i+j)%3]];
                        minus=minus*e[xx[2-j]][yy[(i+j)%3]];
                }
                det=det+plus-minus;
        }
        return det;
}

double Matrix4::Determinant()
{
        double det;
        int i,j;
        det=0;
        for (i=0;i<4;i++)
                for (j=0;j<4;j++)
                        det=det+e[i][j]*Cofactor(i,j);
        return det;
}

Matrix4 Matrix4::Adjoint()
{
        Matrix4 m;
        int i,j;
        for (i=0;i<4;i++)
                for (j=0;j<4;j++)
                        m.e[i][j]=Cofactor(i,j);
        m=m.Transpose();
        return m;
}

Matrix4 Matrix4::Invert()
{
        Matrix4 m;
        double det;
        det=Determinant();
        assert(det!=0.0);
        m=Adjoint() / det;
        return m;
}

void Matrix4::CreateCameraMatrix(Vector3D eye, Vector3D point, Vector3D up)
{
        Vector3D F,U,S;

        F=point-eye;
        F.Normalize();
        U=up;
        U.Normalize();
        S=F.CrossProduct(U);
        U=S.CrossProduct(F);
        e[0][0]=S.x;
        e[0][1]=S.y;
        e[0][2]=S.z;
        e[0][3]=0.0;
        e[1][0]=U.x;
        e[1][1]=U.y;
        e[1][2]=U.z;
        e[1][3]=0.0;
        e[2][0]=-F.x;
        e[2][1]=-F.y;
        e[2][2]=-F.z;
        e[2][3]=0.0;
        e[3][0]=0.0;
        e[3][1]=0.0;
        e[3][2]=0.0;
        e[3][3]=1.0;
}

void Matrix4::AddColumns(int c1, int c2, double *a)
{
        a[0]=e[c1][0]+e[c2][0];
        a[1]=e[c1][1]+e[c2][1];
        a[2]=e[c1][2]+e[c2][2];
        a[3]=e[c1][3]+e[c2][3];
}

void Matrix4::SubColumns(int c1, int c2, double *a)
{
        a[0]=e[c1][0]-e[c2][0];
        a[1]=e[c1][1]-e[c2][1];
        a[2]=e[c1][2]-e[c2][2];
        a[3]=e[c1][3]-e[c2][3];
}


// *** GL STUFF ***

//  m1 m5 (...)
//  m2
//  m3
//  m4

void Matrix4::glGetProjectionMatrix()
{
        double a[16];
        int i,j;
        glGetDoublev( GL_PROJECTION_MATRIX, a);
        for (i=0;i<4;i++)
                for (j=0;j<4;j++)
                        e[i][j]=a[j*4+i];
}

void Matrix4::glGetModelviewMatrix()
{
        double a[16];
        int i,j;
        glGetDoublev( GL_MODELVIEW_MATRIX, a);
        for (i=0;i<4;i++)
                for (j=0;j<4;j++)
                        e[i][j]=a[j*4+i];
}

void Matrix4::glLoadMatrix()
{
        double a[16];
        int i,j;
        for (i=0;i<4;i++)
                for (j=0;j<4;j++)
                        a[j*4+i]=e[i][j];
        glLoadMatrixd(a);
}

void Matrix4::glMultMatrix()
{
        double a[16];
        int i,j;
        for (i=0;i<4;i++)
                for (j=0;j<4;j++)
                        a[j*4+i]=e[i][j];
        glMultMatrixd(a);
}

void Matrix4::Negate()
{
        int i,j;
        for (i=0;i<4;i++)
                for (j=0;j<4;j++)
                        e[i][j]=-e[i][j];
}

