#include <GL/gl.h>
#include <assert.h>
#include <cmath>
#include <iostream>
#include "mesh.h"

Mesh::Mesh (ifstream *f)
{
        string s;
        Vector3D v;
        int i;
        double _u,_v;

        (*f)>>s;
        if (s!="mesh")
        {
                cout <<"error: "<<s<<" expected: mesh\n";
                exit (-1);
        }
        (*f)>>s;
        name=s;

        (*f)>>s;
        if (s!="Loc")
        {
                cout <<"error: "<<s<<" expected: Loc\n";
                exit (-1);
        }
        Loc=Vector3D(f);

        (*f)>>s;
        if (s!="Rot")
        {
                cout <<"error: "<<s<<" expected: Rot\n";
                exit (-1);
        }
        Rot=Vector3D(f);
        Rot.ConvertRadToDeg ();

        (*f)>>s;
        if (s!="Size")
        {
                cout <<"error: "<<s<<" expected: Size\n";
                exit (-1);
        }
        Size=Vector3D(f);

        (*f)>>s;
        if (s!="vertices")
        {
                cout <<"error: "<<s<<" expected: vertices\n";
                exit (-1);
        }

        (*f)>>npoints;

        (*f)>>s;
        if (s!="faces")
        {
                cout <<"error: "<<s<<" expected: faces\n";
                exit (-1);
        }
        (*f)>>nfaces;

        points = new Point3D *[npoints];

        for (i = 0; i < npoints; i++)
        {
                v=Vector3D(f);
                points[i] = new Point3D (v, 0, 0);
                v=Vector3D(f);
                points[i]->SetNormal (v);
                (*f)>>_u;
                (*f)>>_v;
                points[i]->SetTexCoord(_u,_v);
        }

        faces = new Face *[nfaces];
        LoadFaces (f);
        FindBoundingSphereRadius();
}

void Mesh::GenerateList()
{
        int i;
        assert(npoints>0);
        assert(nfaces>0);
        list=glGenLists(1);
        assert(list>0);
        glNewList(list,GL_COMPILE);
        glPushMatrix();
        glBegin (GL_TRIANGLES);
        for (i = 0; i < nfaces; i++)
                faces[i]->Draw ();
        glEnd ();
        glPopMatrix();
        glEndList();
}

Mesh::~Mesh ()
{
        int i;
        for (i = 0; i < npoints; i++)
                delete points[i];
        delete points;
        for (i = 0; i < nfaces; i++)
                delete faces[i];
        delete faces;
}

void Mesh::Draw ()
{
        glCallList(list);
}

void Mesh::DrawNormals ()
{
        int i;
        glPushMatrix ();
        glBegin (GL_LINES);
        for (i = 0; i < npoints; i++)
                points[i]->DrawNormal();
        glEnd ();
        glPopMatrix ();
}

void Mesh::CalcNormals ()
{
        int i;
        for (i = 0; i < nfaces; i++)
                faces[i]->CalcNormal ();
        for (i = 0; i < npoints; i++)
                points[i]->CalcNormal ();
}

string Mesh::GetName ()
{
        return name;
}

Point3D *Mesh::RandomPoint ()
{
        int n;
        if (npoints==0)
                return NULL;
        n = rand () % npoints;
        return points[n];
}

void Mesh::LoadFaces (ifstream * f)
{
        int i;
        for (i = 0; i < nfaces; i++)
                faces[i] = new Face (points, f);
}


void Mesh::GetMinMax (Vector3D * rmin, Vector3D * rmax)
{
        int i;
        *rmin = points[0]->GetPosition ();
        *rmax = *rmin;
        for (i = 0; i < npoints; i++)
        {
                points[i]->GetPosition ().SetMinMax (rmin, rmax);
        }
}

Vector3D Mesh::FindCenter()
{
        Vector3D min,max;
        Vector3D center;
        GetMinMax(&min,&max);
        center=(min+max)/2.0;
        return center;
}

void Mesh::FindBoundingSphereRadius()
{
        Vector3D position;
        int i;
        float d,max;
        max=0.0;
        for (i=0;i<npoints;i++)
        {
                position=points[i]->GetPosition();
                d=Vector3D(0,0,0).Distance(&position);
                if (d>max)
                        max=d;
        }
        BoundingSphereRadius=max;
}

#define TreeSize 100.0

void Mesh::MakeTree()
{
        list=glGenLists(1);
        assert(list>0);
        glNewList(list,GL_COMPILE);
        glBegin(GL_QUADS);

        glTexCoord2f(0.0,0.0);
        glVertex3f(-TreeSize*0.5,TreeSize,0.0);
        glTexCoord2f(0.0,1.0);
        glVertex3f(-TreeSize*0.5,0.0,0.0);
        glTexCoord2f(1.0,1.0);
        glVertex3f(TreeSize*0.5,0.0,0.0);
        glTexCoord2f(1.0,0.0);
        glVertex3f(TreeSize*0.5,TreeSize,0.0);

        glTexCoord2f(0.0,0.0);
        glVertex3f(0.0,TreeSize,-TreeSize*0.5);
        glTexCoord2f(0.0,1.0);
        glVertex3f(0.0,0.0,-TreeSize*0.5);
        glTexCoord2f(1.0,1.0);
        glVertex3f(0.0,0.0,TreeSize*0.5);
        glTexCoord2f(1.0,0.0);
        glVertex3f(0.0,TreeSize,TreeSize*0.5);

        glEnd();
        glEndList();
}

Vector3D Mesh::GetLoc()
{
        return Loc;
}

int Mesh::CheckCrash(Vector3D ContactPoint)
{
        if (ContactPoint.Length()>BoundingSphereRadius)
                return 0;
        cout <<"Mesh "<<name<<" crash detected!!!\n";
        return 1;
}

double Mesh::GetBoundingSphereRadius()
{
        return BoundingSphereRadius;
}
