#include "patch.h"
#include <iostream>
#include <cmath>

Patch::Patch(double _x1, double _z1, double _x2, double _z2, int PatchLevel, HeightMap *_hm, BTTPool *bttpool, int TextureSize, IndicesTree *lit, IndicesTree *rit, GLuint *_indices)
{
        x1=_x1;
        z1=_z1;
        x2=_x2;
        z2=_z2;

        PatchSizeInPoints=(int)pow(2.0,PatchLevel/2.0);
        NumberOfPoints=PatchSizeInPoints+1;

        hm=_hm;

        CreateArrays(TextureSize);

        LeftTree=bttpool->AllocateNode(PatchLevel-1, NULL);
        RightTree=bttpool->AllocateNode(PatchLevel-1, NULL);

        LeftTree->BaseNeighbour=RightTree;
        RightTree->BaseNeighbour=LeftTree;

        LeftIndicesTree=lit;
        RightIndicesTree=rit;

        LeftInfoTree=new TriangleInfoTree(x2,z2,x1,z1,x1,z2,PatchLevel,hm,Vertices,x1,z1, LeftIndicesTree);
        RightInfoTree=new TriangleInfoTree(x1,z1,x2,z2,x2,z1,PatchLevel,hm,Vertices,x1,z1, RightIndicesTree);

        FindMinMaxY();
        indices=_indices;

};

Patch::~Patch()
{
        delete LeftInfoTree;
        delete RightInfoTree;
        delete [] Vertices;
        delete [] TexCoords;
}

void Patch::CreateArrays(int TextureSize)
{
        int i,j;
        double StepX,StepZ;
        double a,b;

        Vertices=new VertexStruct[(NumberOfPoints)*(NumberOfPoints)];
        TexCoords=new TextureStruct[(NumberOfPoints)*(NumberOfPoints)];

        ArraysSize=(NumberOfPoints)*(NumberOfPoints);

        StepX=(x2-x1)/(PatchSizeInPoints);
        StepZ=(z2-z1)/(PatchSizeInPoints);

        a=(1.0-2.0/TextureSize)/PatchSizeInPoints;
        b=1.0/TextureSize;

        for (i=0;i<NumberOfPoints;i++)
        {
                for (j=0;j<NumberOfPoints;j++)
                {
                        Vertices[i+j*(NumberOfPoints)].x=i*StepX;
                        Vertices[i+j*(NumberOfPoints)].y=hm->GetHeight(x1+i*StepX,z1+j*StepZ);
                        Vertices[i+j*(NumberOfPoints)].z=j*StepZ;
                        TexCoords[i+j*(NumberOfPoints)].v=a*i+b;
                        TexCoords[i+j*(NumberOfPoints)].u=a*j+b;
                }
        }
}
void Patch::RenderTree(BTTNode *tree, int node)
{
        if (tree->LeftChild->LeftChild)
                RenderTree(tree->LeftChild, node<<1);
        else
        {
                node=node<<1;
                node=node-1;
                indices[nindices]=CurrentIndices[node][0];
                indices[nindices+1]=CurrentIndices[node][1];
                indices[nindices+2]=CurrentIndices[node][2];
                nindices+=3;
                node=node+1;
                node=node>>1;
        }
        if (tree->RightChild->LeftChild)
                RenderTree(tree->RightChild,(node<<1)+1);
        else
        {
                node=node<<1;
                indices[nindices]=CurrentIndices[node][0];
                indices[nindices+1]=CurrentIndices[node][1];
                indices[nindices+2]=CurrentIndices[node][2];
                nindices+=3;
                node=node>>1;
        }
}

int Patch::GenerateIndices()
{
        nindices=0;
        CurrentInfoTree=LeftInfoTree;
        CurrentIndices=LeftIndicesTree->Get();
        if (LeftTree->LeftChild)
                RenderTree(LeftTree,1);
        else
        {
                indices[nindices]=CurrentIndices[0][0];
                indices[nindices+1]=CurrentIndices[0][1];
                indices[nindices+2]=CurrentIndices[0][2];
                nindices+=3;
        }
        CurrentInfoTree=RightInfoTree;
        CurrentIndices=RightIndicesTree->Get();
        if (RightTree->LeftChild)
                RenderTree(RightTree,1);
        else
        {
                indices[nindices]=CurrentIndices[0][0];
                indices[nindices+1]=CurrentIndices[0][1];
                indices[nindices+2]=CurrentIndices[0][2];
                nindices+=3;
        }
        return nindices;
}

void Patch::Draw(GLuint primitive)
{
        glColor3f(1,1,1);
        glVertexPointer(3,GL_FLOAT,0,Vertices);
        glTexCoordPointer(2,GL_FLOAT,0,TexCoords);
        glBindTexture(GL_TEXTURE_2D,texture);
        glDrawElements(primitive,nindices, GL_UNSIGNED_INT, indices);
}

void Patch::SetNeighbours(Patch *ln, Patch *rn, Patch *tn, Patch *bn)
{
        assert(ln);
        assert(tn);
        assert(rn);
        assert(bn);
        LeftTree->RightNeighbour=ln->RightTree;
        RightTree->LeftNeighbour=tn->LeftTree;
        RightTree->RightNeighbour=rn->LeftTree;
        LeftTree->LeftNeighbour=bn->RightTree;
}

void Patch::FillQs(SplitQueue *Qs)
{
        LandscapeTriangle *t;
        Vector3D v;
        v=Vector3D(0,0,0);
        t=Qs->AllocateTriangle(LeftTree,LeftInfoTree,1);
        Qs->Add(t, &v);
        t=Qs->AllocateTriangle(RightTree,RightInfoTree,1);
        Qs->Add(t, &v);
}

void Patch::FindMinMaxY()
{
        double x,z;
        int i,j;
        double h;
        double StepX,StepZ;
        h=hm->GetHeight(x1,z1);
        MinY=h;
        MaxY=h;
        StepX=(x2-x1)/PatchSizeInPoints;
        StepZ=(z2-z1)/PatchSizeInPoints;
        z=z1;
        for (i=0;i<NumberOfPoints;i++)
        {
                x=x1;
                for (j=0;j<NumberOfPoints;j++)
                {
                        h=hm->GetHeight(x,z);
                        if (h>MaxY)
                                MaxY=h;
                        if (h<MinY)
                                MinY=h;
                        x+=StepX;
                }
                z+=StepZ;
        }
}

int Patch::CheckIfEnabled(double x, double z, Frustum *frustum)
{
        Box box;
        box=Box(x,MinY,z,x+(x2-x1),MaxY,z+(z2-z1));
        if (frustum->BoxInside(&box))
        {
                LeftInfoTree->Visible=1;
                RightInfoTree->Visible=1;
                return 1;
        }
        return 0;
}

void Patch::SetTexture(GLuint _txt)
{
        texture=_txt;
}

void Patch::Disable()
{
        LeftInfoTree->Visible=0;
        RightInfoTree->Visible=0;
}
